| Copyright | (c) Levent Erkok |
|---|---|
| License | BSD3 |
| Maintainer | erkokl@gmail.com |
| Stability | experimental |
| Safe Haskell | None |
| Language | Haskell2010 |
Documentation.SBV.Examples.TP.PigeonHole
Description
Proves the pigeon-hole principle. If a list of integers sum to more than the length
of the list itself, then some cell must contain a value larger than 1.